#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import os
from concurrent import futures
import time
import threading
import random

import grpc

import nudnik
import nudnik.utils as utils

_ONE_DAY_IN_SECONDS = 60 * 60 * 24

class ParseService(nudnik.entity_pb2_grpc.ParserServicer):

    def __init__(self, cfg, metrics):
        super(ParseService, self).__init__()
        self.cfg = cfg
        self.log = utils.get_logger(cfg.debug)
        self.metrics = metrics

    def parse(self, request, context):

        # Generate fake load for incoming request
        for load in self.cfg.load_list:
            utils.generate_load(self.log, load)

        recieved_at = utils.time_ns()

        cdelta = utils.diff_nanoseconds(request.ctime, recieved_at)

        # Calculate retransimt-delta
        if request.rtime != 0:
            rdelta = utils.diff_nanoseconds(request.rtime, recieved_at)
        else:
            rdelta = 0

        if self.cfg.name_mismatch_error is True and not request.name.startswith(self.cfg.name):
            raise NameMismatchException('Client name "{}" does not match server prefix "{}"'.format(request.name, self.cfg.name))

        if (self.metrics.get_fail_ratio() >= self.cfg.fail_ratio):
            self.metrics.add_success()
            status_code = 'OK'
        else:
            self.metrics.add_failure()
            status_code = 'SERVER_ERROR'


        response = {'status_code': status_code, 'ptime': utils.time_ns()}
        grpc_response = nudnik.entity_pb2.Response(**response)

        stat = nudnik.metrics.Stat(request, grpc_response, recieved_at)
        self.metrics.append(stat)

        if self.cfg.chaos > 0 and random.randint(0, self.cfg.cycle_per_hour) <= self.cfg.chaos:
            chaos_exception = utils.ChaosException(self.cfg.chaos_string)
            self.log.fatal(chaos_exception)
            raise chaos_exception

        return grpc_response

    def start_server(self):
        max_workers = max(1, (os.sysconf('SC_NPROCESSORS_ONLN') - 1))
#        max_workers = os.sysconf('SC_NPROCESSORS_ONLN') * 2
        parse_server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
        nudnik.entity_pb2_grpc.add_ParserServicer_to_server(ParseService(self.cfg, self.metrics),parse_server)
        parse_server.add_insecure_port('[::]:{}'.format(self.cfg.port))
        # Non blocking
        parse_server.start()
        self.log.info('Parser Server binded to port {}'.format(self.cfg.port))

        try:
            while True:
                time.sleep(_ONE_DAY_IN_SECONDS)
        except KeyboardInterrupt:
            parse_server.stop(0)
            self.log.warn('Interrupted by user')



#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import time
import threading
import os
import random

import grpc

import nudnik
import nudnik.utils as utils

class ParserClient(object):

    def __init__(self, host, port):
        self.host = host
        self.port = port
# TODO gRPC does not honor values bigger than 4194304
#        max_message_size = (100 * 1024 * 1024)
#        options=[('grpc.max_message_length', -1), ('grpc.max_recieve_message_length', -1), ('grpc.max_send_message_length', -1)]
        options=[]
        self.channel = grpc.insecure_channel('{}:{}'.format(self.host, self.port), options=options)

        # bind the client to the server channel
        self.stub = nudnik.entity_pb2_grpc.ParserStub(self.channel)

    def get_response_for_request(self, request):
        return self.stub.parse(request)

class Stream(threading.Thread):
    def __init__(self, cfg, stream_id, metrics):
        threading.Thread.__init__(self)

        self.cfg = cfg
        self.log = utils.get_logger(cfg.debug)
        self.gtfo = False
        self.stream_id = stream_id
        self.metrics = metrics
        self.name = '{}-{}'.format(cfg.name, stream_id)
        self.log.debug('Stream {} initiated'.format(self.name))

    def run(self):
        self.log.debug('Stream {} started, sending {} messages per second'.format(self.name, (self.cfg.interval * self.cfg.rate)))

        generator_sequence_number = 0
        while not self.gtfo:
            time_start = utils.time_ns()

            try:
                fg = MessageGenerator(self.cfg, self.log, self.name, self.stream_id, generator_sequence_number, self.metrics)
                fg.start()
            except Exception as e:
                self.log.fatal(e)
                self.exit()
            generator_sequence_number += 1

            if self.cfg.chaos > 0 and random.randint(0, self.cfg.cycle_per_hour) <= self.cfg.chaos:
                chaos_exception = utils.ChaosException(self.cfg.chaos_string)
                self.log.fatal(chaos_exception)
                self.exit()
                raise chaos_exception

            elapsed = utils.diff_seconds(time_start, utils.time_ns())
            if elapsed < self.cfg.interval:
                time.sleep(self.cfg.interval - elapsed)

    def exit(self):
        self.gtfo = 1

class MessageGenerator(threading.Thread):

    def __init__(self, cfg, log, name, stream_id, generator_sequence_number, metrics):
        threading.Thread.__init__(self)

        self.cfg = cfg
        self.log = log
        self.stream_id = stream_id
        self.generator_sequence_number = generator_sequence_number
        self.metrics = metrics
        self.started_at = utils.time_ns()
        self.name = name

    def run(self):
        if self.cfg.vv:
            self.log.debug('MessageGenerator {} initiated, sending {} messages'.format(self.name, self.cfg.rate))
        time_start = utils.time_ns()

        client = ParserClient(self.cfg.host, self.cfg.port)

        for index in range(1, self.cfg.rate + 1):
            if 'meta_filepath' in self.cfg and self.cfg.meta_filepath is not None and self.cfg.meta_filepath in ['random', 'urandom', '/dev/random', '/dev/urandom']:
                self.cfg.meta = os.urandom(_MAX_MESSAGE_SIZE_GRPC_BUG)

            request = nudnik.entity_pb2.Request(name=self.name,
                                     stream_id=self.stream_id,
#                                     sequence_id=self.generator_sequence_number,
                                     message_id=(self.generator_sequence_number * self.cfg.rate) + index,
                                     ctime=utils.time_ns(),
                                     meta=self.cfg.meta,
                                     load=self.cfg.load_list)

            for load in request.load:
                utils.generate_load(self.log, load)

            retry_count = 0
            try_count = 1 + self.cfg.retry_count
            send_was_successful = False
            while (not send_was_successful and ((self.cfg.retry_count < 0) or (try_count > 0))):
                response = client.get_response_for_request(request)
                recieved_at = utils.time_ns()

                send_was_successful = ((response.status_code == 0) and (self.metrics.get_fail_ratio() >= self.cfg.fail_ratio))

                if send_was_successful:
                    self.metrics.add_success()
                    stat = nudnik.metrics.Stat(request, response, recieved_at)
                    self.metrics.append(stat)
                else:
                    self.metrics.add_failure()
                    try_count -= 1
                    retry_count += 1
                    request.rtime=utils.time_ns()
                    request.rcount = retry_count

        elapsed = utils.diff_seconds(time_start, utils.time_ns())
        if elapsed > self.cfg.interval:
            self.log.warn('cdelta {} for rate {} exceeds interval {}'.format(elapsed, self.cfg.rate, self.cfg.interval))




#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import time
import threading
import requests_unixsocket

import nudnik.utils as utils

class Metrics(threading.Thread):

    def __init__(self, cfg):
        threading.Thread.__init__(self)
        self.gtfo = False
        self.lock = threading.Lock()
        self.metrics = list()
        self.successful_requests = 0
        self.failed_requests = 0
        self.cfg = cfg
        self.log = utils.get_logger(cfg.debug)
        self.workers = list()

        if 'file' in self.cfg.metrics:
            self.file_path = cfg.file_path
        if 'influxdb' in self.cfg.metrics:
            if cfg.influxdb_protocol == 'http+unix':
                self.influxdb_host = cfg.influxdb_socket_path.replace('/', '%2F')
            else:
                self.influxdb_host = cfg.influxdb_host

            create_influxdb_database(self, self.log, cfg.influxdb_protocol, self.influxdb_host, cfg.influxdb_database_name)
            self.influxdb_url = cfg.influxdb_url.format(influxdb_protocol=cfg.influxdb_protocol,
                                                        influxdb_host=self.influxdb_host,
                                                        influxdb_database_name=cfg.influxdb_database_name)

        self.log.debug('Metrics thread initiated')

    def run(self):
        self.log.debug('Running {}'.format(self.name))
        while not self.gtfo:
            if self.cfg.vv:
                self.log.debug('Reporting {} items'.format(len(self.metrics)))

            time_start = utils.time_ns()

            current_report = list(self.metrics)
            current_report_length = len(current_report)
            if len(current_report) > 1:
                if self.cfg.vvv:
                    self.log.debug('Reporting {}/{} items'.format(current_report_length, len(self.metrics)))


                if self.cfg.debug:
                    for stat in _parse_stats(current_report, self.cfg.out_format, self.cfg.out_retransmit_format):
                        self.log.debug(stat)
                elif 'stdout' in self.cfg.metrics:
                    for stat in _parse_stats(current_report, self.cfg.out_format, self.cfg.out_retransmit_format):
                        self.log.info(stat)

                if 'file' in self.cfg.metrics:
                    thread = FileMetrics(self.log, self.file_path, current_report, self.cfg.out_format, self.cfg.out_retransmit_format)
                    thread.start()
                    self.workers.append(thread)
                if 'influxdb' in self.cfg.metrics:
                    thread = InfluxdbMetrics(self.log, self.influxdb_url, current_report, self.cfg.influxdb_format, self.cfg.influxdb_retransmit_format)
                    thread.start()
                    self.workers.append(thread)

                for i in range(0, current_report_length):
                    self.metrics.pop(0)

                while len(self.workers) > 0:
                    for index, thread in enumerate(self.workers):
                        thread.join()
                        self.workers.pop(index)

            elapsed = utils.diff_seconds(time_start, utils.time_ns())
            if elapsed < self.cfg.interval:
                time.sleep(self.cfg.interval - elapsed)

    def add_success(self):
        with self.lock:
            self.successful_requests += 1

    def add_failure(self):
        with self.lock:
            self.failed_requests += 1

    def get_fail_ratio(self):
        total = self.failed_requests + self.successful_requests
        try:
            current_fail_ratio = float((self.failed_requests / total) * 100)
        except ZeroDivisionError:
            current_fail_ratio = 100.0

        if self.cfg.vvv:
            logformat = 'failed={},success={},current_fail_ratio={},conf_fail_ratio={}'
            self.log.debug(logformat.format(self.failed_requests,
                                            self.successful_requests,
                                            current_fail_ratio,
                                            self.cfg.fail_ratio))
        return current_fail_ratio

    def append(self, stat):
        self.metrics.append(stat)

    def exit(self):
        self.gtfo = 1

class FileMetrics(threading.Thread):
    def __init__(self, log, path, data, format, retransmit_format):
        threading.Thread.__init__(self)
        self.log = log
        self.path = path
        self.data = data
        self.format = format
        self.retransmit_format = retransmit_format

    def run(self):
        data = list()
        for stat in _parse_stats(self.data, self.format, self.retransmit_format):
            data.append(stat)
        self.log.debug('Writing {} items to {}'.format(len(data), self.path))
        with open(self.path, 'a') as metricsfile:
            metricsfile.write('\n'.join(data))
            metricsfile.write('\n')
        return True

class InfluxdbMetrics(threading.Thread):
    def __init__(self, log, url, data, format, retransmit_format):
        threading.Thread.__init__(self)
        self.log = log
        self.session = requests_unixsocket.Session()
        self.url = url
        self.data = data
        self.format = format
        self.retransmit_format = retransmit_format

    def run(self):
        data = list()
        for stat in _parse_stats(self.data, self.format, self.retransmit_format):
            data.append(stat)

        self.log.debug('Writing {} items to InfluxDB'.format(len(data)))
        session = requests_unixsocket.Session()
        res = session.post(self.url, data='\n'.join(data))
        session.close()
        if res.status_code != 204:
            self.log.error('Response: "{}"'.format(res.text))
            return False
        return True

def create_influxdb_database(self, log, protocol, host, database_name):
    # https://docs.influxdata.com/influxdb/v1.7/tools/api/
    # https://docs.influxdata.com/influxdb/v1.7/query_language/database_management/#create-database
    query ='q=CREATE DATABASE "{}"'.format(database_name)
    session = requests_unixsocket.Session()
    res = session.post('{}://{}/query?{}'.format(protocol, host, query))
    session.close()
    if res.status_code == 200:
        log.debug('Response to {}: "{}"'.format(query, res.text))
    else:
        log.error('Response to {}: "{}"'.format(query, res.text))
        raise RuntimeException(res.text)

class Stat(object):
    def __init__(self, request, response, recieved_at):
        self.request = request
        self.response = response
        self.recieved_at = recieved_at

def _parse_stats(stats, format, retransmit_format):
    for stat in stats:
        if stat.request.rcount == 0:
            dataformat = format
        else:
            dataformat = retransmit_format

        statstring = dataformat.format(recieved_at_str=str(stat.recieved_at),
                                       recieved_at=stat.recieved_at,
                                       status_code=stat.response.status_code,
                                       req=stat.request,
                                       cdelta=utils.diff_nanoseconds(stat.request.ctime, stat.response.ptime),
                                       rdelta=utils.diff_nanoseconds(stat.request.rtime, stat.response.ptime),
                                       rtt=utils.diff_nanoseconds(stat.request.ctime, stat.recieved_at))
        yield statstring




#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys

# Because of protobuf issue #1491 :(
sys.path.append(os.path.dirname(__file__))

import nudnik.entity_pb2
import nudnik.entity_pb2_grpc

__version__ = '0.0.12'





# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import entity_pb2 as entity__pb2


class ParserStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.parse = channel.unary_unary(
        '/nudnik.entity.Parser/parse',
        request_serializer=entity__pb2.Request.SerializeToString,
        response_deserializer=entity__pb2.Response.FromString,
        )


class ParserServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def parse(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ParserServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'parse': grpc.unary_unary_rpc_method_handler(
          servicer.parse,
          request_deserializer=entity__pb2.Request.FromString,
          response_serializer=entity__pb2.Response.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'nudnik.entity.Parser', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: entity.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='entity.proto',
  package='nudnik.entity',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0c\x65ntity.proto\x12\rnudnik.entity\"w\n\x04Load\x12/\n\tload_type\x18\x01 \x01(\x0e\x32\x1c.nudnik.entity.Load.LoadType\x12\r\n\x05value\x18\x02 \x01(\t\"/\n\x08LoadType\x12\x07\n\x03rtt\x10\x00\x12\x08\n\x04rttr\x10\x01\x12\x07\n\x03\x63pu\x10\x02\x12\x07\n\x03mem\x10\x03\"\xb2\x01\n\x07Request\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tstream_id\x18\x02 \x01(\x05\x12\x13\n\x0bsequence_id\x18\x03 \x01(\x05\x12\x12\n\nmessage_id\x18\x04 \x01(\x05\x12\r\n\x05\x63time\x18\x05 \x01(\x03\x12\r\n\x05rtime\x18\x06 \x01(\x03\x12\x0e\n\x06rcount\x18\x07 \x01(\x05\x12\x0c\n\x04meta\x18\x08 \x01(\x0c\x12!\n\x04load\x18\t \x03(\x0b\x32\x13.nudnik.entity.Load\"\x8c\x01\n\x08Response\x12\x37\n\x0bstatus_code\x18\x01 \x01(\x0e\x32\".nudnik.entity.Response.StatusCode\x12\r\n\x05ptime\x18\x02 \x01(\x03\"8\n\nStatusCode\x12\x06\n\x02OK\x10\x00\x12\x10\n\x0cSERVER_ERROR\x10\x01\x12\x10\n\x0c\x43LIENT_ERROR\x10\x02\x32\x44\n\x06Parser\x12:\n\x05parse\x12\x16.nudnik.entity.Request\x1a\x17.nudnik.entity.Response\"\x00\x62\x06proto3')
)



_LOAD_LOADTYPE = _descriptor.EnumDescriptor(
  name='LoadType',
  full_name='nudnik.entity.Load.LoadType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='rtt', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='rttr', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='cpu', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='mem', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=103,
  serialized_end=150,
)
_sym_db.RegisterEnumDescriptor(_LOAD_LOADTYPE)

_RESPONSE_STATUSCODE = _descriptor.EnumDescriptor(
  name='StatusCode',
  full_name='nudnik.entity.Response.StatusCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SERVER_ERROR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIENT_ERROR', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=418,
  serialized_end=474,
)
_sym_db.RegisterEnumDescriptor(_RESPONSE_STATUSCODE)


_LOAD = _descriptor.Descriptor(
  name='Load',
  full_name='nudnik.entity.Load',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='load_type', full_name='nudnik.entity.Load.load_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nudnik.entity.Load.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LOAD_LOADTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31,
  serialized_end=150,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='nudnik.entity.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='nudnik.entity.Request.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stream_id', full_name='nudnik.entity.Request.stream_id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sequence_id', full_name='nudnik.entity.Request.sequence_id', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_id', full_name='nudnik.entity.Request.message_id', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ctime', full_name='nudnik.entity.Request.ctime', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rtime', full_name='nudnik.entity.Request.rtime', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rcount', full_name='nudnik.entity.Request.rcount', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='meta', full_name='nudnik.entity.Request.meta', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='load', full_name='nudnik.entity.Request.load', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=153,
  serialized_end=331,
)


_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='nudnik.entity.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status_code', full_name='nudnik.entity.Response.status_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ptime', full_name='nudnik.entity.Response.ptime', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RESPONSE_STATUSCODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=334,
  serialized_end=474,
)

_LOAD.fields_by_name['load_type'].enum_type = _LOAD_LOADTYPE
_LOAD_LOADTYPE.containing_type = _LOAD
_REQUEST.fields_by_name['load'].message_type = _LOAD
_RESPONSE.fields_by_name['status_code'].enum_type = _RESPONSE_STATUSCODE
_RESPONSE_STATUSCODE.containing_type = _RESPONSE
DESCRIPTOR.message_types_by_name['Load'] = _LOAD
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Load = _reflection.GeneratedProtocolMessageType('Load', (_message.Message,), dict(
  DESCRIPTOR = _LOAD,
  __module__ = 'entity_pb2'
  # @@protoc_insertion_point(class_scope:nudnik.entity.Load)
  ))
_sym_db.RegisterMessage(Load)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), dict(
  DESCRIPTOR = _REQUEST,
  __module__ = 'entity_pb2'
  # @@protoc_insertion_point(class_scope:nudnik.entity.Request)
  ))
_sym_db.RegisterMessage(Request)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), dict(
  DESCRIPTOR = _RESPONSE,
  __module__ = 'entity_pb2'
  # @@protoc_insertion_point(class_scope:nudnik.entity.Response)
  ))
_sym_db.RegisterMessage(Response)



_PARSER = _descriptor.ServiceDescriptor(
  name='Parser',
  full_name='nudnik.entity.Parser',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=476,
  serialized_end=544,
  methods=[
  _descriptor.MethodDescriptor(
    name='parse',
    full_name='nudnik.entity.Parser.parse',
    index=0,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_PARSER)

DESCRIPTOR.services_by_name['Parser'] = _PARSER

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import sys

import nudnik.metrics
import nudnik.server
import nudnik.client
import nudnik.utils as utils

def main():
    args = utils.parse_args()
    cfg = utils.parse_config(args)
    log = utils.get_logger(cfg.debug)

    metricsthread = nudnik.metrics.Metrics(cfg)
    metricsthread.start()

    if cfg.server:
        log.debug('Running Nudnik in server mode')
        server = nudnik.server.ParseService(cfg, metricsthread)
        server.start_server()
    else:
        log.debug('Running Nudnik in client mode')
        streams = list()
        log.debug('Starting {} streams'.format(cfg.streams))
        for i in range(1, cfg.streams + 1):
            stream = nudnik.client.Stream(cfg, i, metricsthread)
            streams.append(stream)
            stream.start()

        while len(streams) > 0:
            for index, stream in enumerate(streams):
                try:
                    if stream.gtfo:
                        streams.pop(index)
                    else:
                        stream.join(0.25)
                except KeyboardInterrupt:
                    for stream in streams:
                        stream.exit()
                        streams.pop(index)

    metricsthread.exit()
    log.debug('You are the weakest link, goodbye!'.format(''))
    return 1

if __name__ == "__main__":
    sys.exit(main())




#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import os
from setuptools import setup

import setuptools

import re

def version():
    INITFILE="nudnik/__init__.py"
    initstring = open('nudnik/__init__.py', "rt").read()
    version = re.search(r"^__version__ = '(.*)'", initstring, re.M).group(1)
    return version

with open("README.md", "r") as f:
    long_description = f.read()

with open('requirements.txt') as f:
     install_reqs = f.read().splitlines()

setuptools.setup(
     name='nudnik',
     version=version(),
     url="https://github.com/salosh/nudnik.git",
     scripts=['nudnik/bin/nudnik'],
     author="Salo Shp",
     author_email="support@salosh.org",
     license='GPLv3',
     description="gRPC load-balancing testing tool",
     long_description=long_description,
     long_description_content_type="text/markdown",
     packages=setuptools.find_packages(),
     install_requires=install_reqs,
     use_2to3=True,
     classifiers=[
         "Environment :: Console",
         "Intended Audience :: System Administrators",
         "Programming Language :: Python",
         "Topic :: System :: Distributed Computing",
         "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
         "Natural Language :: English",
         "Operating System :: OS Independent",
     ],
     zip_safe=False,
)



#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
import argparse
import yaml
import os
import sys
import time
import re
from datetime import datetime
import random

if 'FileNotFoundError' not in sys.modules:
    from exceptions import IOError as FileNotFoundError

import nudnik

_BILLION = 10**9

class NudnikConfiguration(dict):
    pass

def parse_args():
    parser = argparse.ArgumentParser(
        description='Nudnik - gRPC load tester',
        epilog='2018 (C) Salo Shp <https://github.com/salosh/nudnik.git>'
    )
    parser.add_argument('--config-file', '-f',
                        type=str,
                        help='Path to YAML config file')
    parser.add_argument('--host', '-H',
                        type=str,
                        help='host')
    parser.add_argument('--port', '-p',
                        type=int,
                        help='port')
    parser.add_argument('--server', '-S',
                        action='store_true',
                        help='Operation mode (default: client)')
    parser.add_argument('--name', '-n',
                        type=str,
                        help='Parser name')
    parser.add_argument('--name-mismatch-error',
                        action='store_true',
                        help='Fail request on name mismatch (default: False)')
    parser.add_argument('--meta', '-M',
                        type=str,
                        help='Send this extra data with every request')
    parser.add_argument('--streams', '-s',
                        type=int,
                        help='Number of streams (Default: 1)')
    parser.add_argument('--interval', '-i',
                        type=int,
                        help='Number of seconds per stream message cycle (Default: 1)')
    parser.add_argument('--rate', '-r',
                        type=int,
                        help='Number of messages per interval (Default: 10)')
    parser.add_argument('--chaos', '-c',
                        type=int,
                        help='Compute statistical process level random crashes [0, 3600/interval] (Default: 0)')
    parser.add_argument('--load', '-l',
                        type=str,
                        nargs=2,
                        action='append',
                        metavar=('load_type', 'load_value'),
                        dest='load',
                        help='Add artificial load [rtt, rttr, cpu, mem] (Default: None)')
    parser.add_argument('--retry-count',
                        type=int,
                        help='Number of times to re-send failed messages (Default: -1, which means infinite times)')
    parser.add_argument('--fail-ratio',
                        type=str,
                        help='Percent of requests to intentionally fail (Default: 0)')
    parser.add_argument('--metrics', '-m',
                        type=str,
                        action='append',
                        choices=['stdout', 'file', 'influxdb'],
                        help='Enable metrics outputs (Default: None)')
    parser.add_argument('--file-path', '-F',
                        type=str,
                        help='Path to exported metrics file (Default: ./nudnikmetrics.out)')
    parser.add_argument('--influxdb-socket-path',
                        type=str,
                        help='Absolute path to InfluxDB Unix socket (Default: /var/run/influxdb/influxdb.sock)')
    parser.add_argument('--influxdb-database-name',
                        type=str,
                        help='InfluxDB database name (Default: nudnikmetrics)')
    parser.add_argument('--debug', '-d',
                        action='store_true',
                        help='Debug mode (default: False)')
    parser.add_argument('--verbose', '-v',
                        action='count',
                        help='Verbose mode, specify multiple times for extra verbosity (default: None)')
    parser.add_argument('--version', '-V',
                        action='version',
                        version='Nudnik v{} - 2018 (C) Salo Shp <https://github.com/salosh/nudnik.git>'.format(nudnik.__version__),
                        help='Display Nudnik version')

    args = parser.parse_args()
    return args

def parse_config(args):
    cfg = NudnikConfiguration()

    DEFAULTS = {
      'config_file': 'config.yml',
      'host': '127.0.0.1',
      'port': 5410,
      'server': False,
      'name': 'NAME',
      'name_mismatch_error': False,
      'meta': None,
      'streams': 1,
      'interval': 1,
      'rate': 1,
      'chaos': 0,
      'chaos_string': 'In all chaos there is a cosmos, in all disorder a secret order. #Carl_Jung_FTW',
      'load': [],
      'retry_count': -1,
      'fail_ratio': 0,
      'metrics': [],
      'file_path': './nudnikmetrics.out',
      'out_format': '{recieved_at_str},{status_code},{req.name},{req.message_id},{req.ctime},{cdelta},rtt={rtt}',
      'out_retransmit_format': '{recieved_at_str},{status_code},{req.name},{req.message_id},{req.ctime},{req.rtime},{cdelta},{rdelta},{req.rcount},rtt={rtt}',
      'influxdb_socket_path': '/var/run/influxdb/influxdb.sock',
      'influxdb_protocol': 'http+unix',
      'influxdb_host': '127.0.0.1:8086',
      'influxdb_database_name': 'nudnikmetrics',
      'influxdb_url': '{influxdb_protocol}://{influxdb_host}/write?db={influxdb_database_name}&precision=ns',
      # First field for InfluxDB is measurment name
      'influxdb_format': 'request,status={status_code},name={req.name} mid={req.message_id},ctime={req.ctime},cdelta={cdelta},rtt={rtt} {recieved_at}',
      'influxdb_retransmit_format': 'request,status={status_code},name={req.name} mid={req.message_id},ctime={req.ctime},rtime={req.rtime},cdelta={cdelta},rdelta={rdelta},rcount={req.rcount},rtt={rtt} {recieved_at}',
      'debug': False,
      'verbose': 0,
    }

    for key in DEFAULTS:
        env_key_name = 'NUDNIK_{key_name}'.format(key_name=key.upper())
        if key in args.__dict__ and vars(args)[key]:
            value = vars(args)[key]
        elif env_key_name in os.environ:
            value = os.environ[env_key_name]
            if isinstance(DEFAULTS[key], int):
                value = int(value)
            elif isinstance(DEFAULTS[key], str):
                value = str(value)
            else:
                value = value
        else:
            value = DEFAULTS[key]

        setattr(cfg, key, value)

    try:
        with open(cfg.config_file, 'r') as ymlfile:
            ymlcfg = yaml.load(ymlfile)

            if not isinstance(ymlcfg, dict):
                raise FileNotFoundError

            for confkey in DEFAULTS:
                if ymlcfg[confkey] is not None and not getattr(cfg, confkey, None):
                    setattr(cfg, confkey, ymlcfg[confkey])

    except yaml.parser.ParserError as e:
        print('Nudnik Configuration error: {}'.format(e))
        sys.exit(1)
    except FileNotFoundError:
        print('Configuration file "{}" was not found, ignoring.'.format(cfg.config_file))
        pass

    cfg.cycle_per_hour = int( 3600 / cfg.interval )

    # Clear '%' sign if provided
    fail_ratio = (re.match(r"(([0-9])*(\.)*([0-9])*)", str(cfg.fail_ratio))).groups()[0]
    cfg.fail_ratio = float(fail_ratio) if (fail_ratio != '') else 0.0

    for i in range(1, 6):
        attr = 'v'*i
        setattr(cfg, attr, cfg.verbose >= i)

    if cfg.v:
        cfg.debug = True

    if cfg.vvvvv:
        for key in DEFAULTS:
            print('{} - {}'.format(key, getattr(cfg, key)))

    _MAX_MESSAGE_SIZE_GRPC_BUG = 4194304 - 48
    if cfg.meta is not None and cfg.meta != '' and cfg.meta[0] == '@':
        cfg.meta_filepath = cfg.meta[1:]

        try:
            with open(cfg.meta_filepath, 'rb') as f:
                cfg.meta = f.read(_MAX_MESSAGE_SIZE_GRPC_BUG)

        except Exception as e:
            print('Could not open meta file "{}", {}'.format(cfg.meta_filepath, str(e)))
            cfg.meta = None
            pass

    load_list = list()
    if cfg.load is not None:
        for load in cfg.load:
            load = nudnik.entity_pb2.Load(load_type=load[0], value=load[1])
            load_list.append(load)
    cfg.load_list = load_list
    return cfg

def get_logger(debug=False):
  ''' Initialize logging configuration '''
  format = '%(asctime)-15s %(levelname).5s %(process)d %(threadName)s [%(module)s] [%(funcName)s:%(lineno)d] - %(message)s'
  logging.basicConfig(format=format)
  logger = logging.getLogger(__name__)

  if debug:
      logger.setLevel(logging.DEBUG)
#      logger.debug('Debug-mode logging enabled')
  else:
      logger.setLevel(logging.INFO)

  return logger

def generate_load(logger, load):
    if load.load_type == 0:
        time_sleep = float(load.value)
        logger.debug('Sleeping for {}'.format(time_sleep))
        time.sleep(time_sleep)
    elif load.load_type == 1:
        time_sleep = random.uniform(0.0, float(load.value))
        logger.debug('Sleeping for random value {}'.format(time_sleep))
        time.sleep(time_sleep)
    elif load.load_type == 2:
        time_load = load.value
        logger.debug('CPU loading for {} seconds'.format(time_load))
        for i in range(0, os.sysconf('SC_NPROCESSORS_ONLN')):
            cpu_load_thread = FakeLoadCpu(time_load).start()
    elif load.load_type == 3:
        amount_in_mb = load.value
        logger.debug('Loading {} MB to RAM'.format(amount_in_mb))
        mem_load_thread = FakeLoadMem(amount_in_mb)
        mem_load_thread.start()

def time_ns():
    # Python < 3.7 doesn't support time.time_ns(), this function unifies metrics measurments
    return int(("%.9f" % time.time()).replace('.',''))

def diff_nanoseconds(before, after):
    return (after - before)

def diff_seconds(before, after):
    return ((after - before) / _BILLION)

def time_to_date(timestamp):
    return datetime.fromtimestamp( timestamp / _BILLION )

class ChaosException(Exception): pass
